from broad_crawl.urls import extract_urls
from file_logger import get_logger

from typing import Dict, Any, List
from datetime import datetime, timezone

logger = get_logger('broad_crawl')


async def broad_crawl(
    # URL of page to crawl.
    url: str,
    # placeholder for crawler_cluster dependency injection of TaskExecuter instance.
    task_executer,
    # placeholder for crawler_cluster dependency injection.
    task_id: int,
    # depth of URL from start URL. If provided URL is a start URL for a new crawl task, depth should be 0.
    page_depth: int = 0,
    # max number of nested pages to follow from start URL.
    max_depth: int = None,
    # domains that are allowed to be crawled.
    allowed_domains: List[str] = None,
    # domains that are prohibited from being crawled.
        prohibited_domains: List[str] = None
) -> Dict[str, Any]:
    """Scrape page at provided URL and create tasks to scrape URls extracted from the page."""
    # use task_executer's members for processing and caching.
    loader = task_executer.loader
    task_queue = task_executer.task_queue
    redis = task_queue.redis
    # fetch page.
    await loader.fetch(url)
    # save page data.
    result = {
        'html': loader.html,
        'url': loader.url,
        'depth': page_depth,
        'time_utc': datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
    }
    logger.info(
        f"[crawl {task_id}] Saving depth {page_depth} page: {loader.url}")
    if page_depth == 0:
        # start URL (i.e. URL with page_depth 0) will not have been added to task's queued URL Set because it was not an extracted URL. add it now so we don't queue it multiple times.
        await redis.sadd(f'{task_id}_queued_urls', url.rstrip(' /'))
    # depth of URLs extracted from page is 1+depth of current page.
    page_urls_depth = page_depth + 1
    if max_depth is None or page_urls_depth <= max_depth:
        # extract page URLs.
        urls = await extract_urls(loader.root, task_id, loader.url, redis, allowed_domains, prohibited_domains)
        # create tasks for all extracted URLs.
        new_tasks = [{
            'function': 'broad_crawl',
            'module': 'broad_crawl',
            'page_depth': page_urls_depth,
            'max_depth': max_depth,
            'task_id': task_id,
            'url': url,
            'allowed_domains': allowed_domains,
            'prohibited_domains': prohibited_domains,
        } for url in urls]
        logger.info(f"Queueing {len(new_tasks)} new tasks for extracted URLs.")
        # queue tasks for all extracted URLs.
        await task_queue.put(new_tasks)
    remaining_tasks_key = f"{task_id}_remaining_tasks"
    # decrement 'remaining URLs' counter for the task just completed.
    await redis.decr(remaining_tasks_key)
    # check if crawl is finshed.
    remaining_tasks_count = await redis.get(remaining_tasks_key)
    if int(remaining_tasks_count.decode()) < 1:
        # crawl is finished. remove temporary keys that were needed during the crawl process.
        await redis.delete(f'{task_id}_queued_urls')
        await redis.delete(remaining_tasks_key)
        return {'is_last': True, 'result': result}
    return {'is_last': False, 'result': result}
