from aioredis import Redis
from urllib.parse import urlsplit, urlunsplit, SplitResult
from typing import List, Tuple
import re


def absolute_url(url: str, split_url: SplitResult, split_page_url: SplitResult) -> Tuple[str, SplitResult]:
    """Convert url to an absolute URL if it is a relative URL.
    page_url should be the URL of the page url was extracted from.
    """
    if split_url.netloc and split_url.scheme == '':
        # URL is already an absolute URL.
        return url, split_url
    # construct absolute URL.
    netloc = split_url.netloc or split_page_url.netloc
    scheme = split_url.scheme or split_page_url.scheme
    url = urlunsplit((scheme, netloc, split_url.path,
                      split_url.query, split_url.fragment))
    return url, urlsplit(url)


async def extract_urls(root, task_id: str, page_url: str,
                       redis: Redis,
                       allowed_domains: List[str] = None,
                       prohibited_domains: List[str] = None) -> List[str]:
    """Extract new/unscraped URLs from current page's HTML."""
    # extract URLs from HTML.
    urls = {str(u).rstrip('/ ') for u in root.xpath('//a[@href]/@href')}
    # cache map of URL to URL split as split URL will be needed a few times.
    urls = {u: urlsplit(u) for u in urls}
    # filter out URLs that are not likely to be crawlable pages.
    urls = {u: usplit for u, usplit in urls.items() if not re.search(
        r'\.(js|css|pdf|csv|tsv|json|apng|avif|gif|jpg|jpeg|jfif|pjpeg|pjp|png|svg|webp|bmp|ico|cur|tif|tiff)$', usplit.netloc)}
    # make sure all URLs are absolute URLs.
    split_page_url = urlsplit(page_url)
    urls = dict(absolute_url(u, usplit, split_page_url)
                for u, usplit in urls.items())
    # filter out URLs that are not from allowed domains.
    if allowed_domains:
        urls = {u: usplit for u, usplit in urls.items(
        ) if usplit.netloc in allowed_domains}
    # filter out URLs that are from prohibited domains.
    if prohibited_domains:
        urls = {u: usplit for u, usplit in urls.items(
        ) if usplit.netloc not in prohibited_domains}
    # filter out URLs that have already been scraped or are currently queued.
    urls = [u for u in urls.keys() if not await redis.sismember(f'{task_id}_queued_urls', u)]
    if len(urls):
        # mark the URLs as seen so we don't process them again.
        await redis.sadd(f'{task_id}_queued_urls', *urls)
        # increment 'remaining URLs' counter for this task by the number of new tasks just created, so we know that this crawl is not finished.
        await redis.incrby(f"{task_id}_remaining_tasks", len(urls))
    return urls
